from dataclasses import dataclass
from typing import List, Optional

from dataclasses_json import dataclass_json

from .badge import Badge
from .fields import default
from .score_stats import ScoreStats


@dataclass_json
@dataclass
class Player:
    """Player data from ScoreSaber"""

    player_id: str = default("playerId")
    player_name: str = default("playerName")
    avatar: str = default()
    rank: int = default()
    country_rank: int = default("countryRank")
    pp: float = default()
    country: str = default()
    role: Optional[str] = default()
    badges: List[Badge] = default()
    history: str = default()
    permissions: int = default()
    inactive: int = default()
    banned: int = default()

    score_stats: Optional[ScoreStats] = default("scoreStats")



import logging
from asyncio import AbstractEventLoop
from typing import *

from faker import Faker
from outcache import CacheAsync

from .http_client import HttpClient
from .models import Player, Score
from .scoresaber_provider import ScoreSaberProvider


class ScoreSaber:
    _url = "https://new.scoresaber.com/api"

    def __init__(self, loop: Optional[AbstractEventLoop] = None, test_mode: bool = False):
        self.log = logging.getLogger(__name__)
        self._http_client = None
        self.loop = loop
        self.test_mode = test_mode

        if test_mode:
            self.faker = Faker()
            Faker.seed(76561198283584459)
            self.faker.add_provider(ScoreSaberProvider)

    async def start(self):
        if self._http_client is None:
            self._http_client = HttpClient(loop=self.loop)

    async def close(self):
        if self._http_client is not None:
            await self._http_client.close()
            self._http_client = None

    @CacheAsync(minutes=2)
    async def _get_player_basic(self, player_id: str) -> Dict:
        if self.test_mode:
            return {"playerInfo": self.faker.player_basic(player_id).to_dict()}

        return await self._http_client.get(f"{self._url}/player/{player_id}/basic")

    async def get_player_basic(self, player_id: str) -> Player:
        response = await self._get_player_basic(player_id)

        return Player.from_dict(response["playerInfo"])

    @CacheAsync(minutes=2)
    async def _get_player_full(self, player_id: str) -> Dict:
        if self.test_mode:
            return {"playerInfo": self.faker.player_full(player_id).to_dict()}

        return await self._http_client.get(f"{self._url}/player/{player_id}/full")

    async def get_player_full(self, player_id: str) -> Player:
        response = await self._get_player_full(player_id)

        return Player.from_dict(response["playerInfo"])

    @CacheAsync(minutes=2)
    async def _get_recent_scores(self, player_id: str, page: int = 1) -> Dict:
        if self.test_mode:
            return {"scores": [score.to_dict() for score in self.faker.scores(8)]}

        return await self._http_client.get(f"{self._url}/player/{player_id}/scores/recent/{page}")

    async def get_recent_scores(self, player_id: str, page: int = 1) -> List[Score]:
        response = await self._get_recent_scores(player_id, page)

        recent_score_list = []

        for recent_score in response["scores"]:
            recent_score_list.append(Score.from_dict(recent_score))

        return recent_score_list

    @CacheAsync(minutes=2)
    async def _get_top_scores(self, player_id: str, page: int = 1) -> Dict:
        if self.test_mode:
            return {"scores": [score.to_dict() for score in self.faker.scores(8)]}

        return await self._http_client.get(f"{self._url}/player/{player_id}/scores/top/{page}")

    async def get_top_scores(self, player_id: str, page: int = 1) -> List[Score]:
        response = await self._get_top_scores(player_id, page)

        top_score_list = []

        for top_score in response["scores"]:
            top_score_list.append(Score.from_dict(top_score))

        return top_score_list

    async def __aenter__(self):
        await self.start()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.close()



class ScoreSaberException(Exception):
    def __init__(self, status: int, url: str) -> None:
        self.status = status
        self.url = url

        super().__init__(f"Score Saber returned {self.status} for {self.url}")


class NotFoundException(ScoreSaberException):
    pass


class ServerException(ScoreSaberException):
    pass



import asyncio
import logging

import aiohttp
from aiohttp import ClientResponse, ClientResponseError

from .errors import ScoreSaberException, NotFoundException, ServerException


class HttpClient(aiohttp.ClientSession):
    RETRIES = 10

    def __init__(self, *args, **kwargs):
        super().__init__(raise_for_status=True, *args, **kwargs)

    async def _request(self, *args, **kwargs) -> ClientResponse:
        retries = 0

        while True:
            try:
                response = await super()._request(*args, **kwargs)
            except ClientResponseError as error:
                if error.status == 404:
                    raise NotFoundException(error.status, str(error.request_info.real_url)) from error
                elif error.status == 500:
                    raise ServerException(error.status, str(error.request_info.real_url)) from error

                raise ScoreSaberException(error.status, str(error.request_info.real_url)) from error

            if response.status == 200:
                return response

            if retries > self.RETRIES:
                raise ScoreSaberException(response.status, str(response.real_url))

            sleep = 2 ** retries

            logging.warning(f"[{retries}/{self.RETRIES}] Ratelimited! Waiting {sleep} seconds...")
            await asyncio.sleep(sleep)

            retries += 1

    async def get(self, url, *args, **kwargs):
        response = await self._request('GET', url, *args, **kwargs)
        return await response.json()



from .base_enum import BaseEnum


class Characteristic(BaseEnum):
    STANDARD = "Standard"
    ONE_SABER = "OneSaber"
    NO_ARROWS = "NoArrows"
    DEGREE_90 = "90Degree"
    DEGREE_360 = "360Degree"
    LIGHTSHOW = "Lightshow"
    LAWLESS = "Lawless"
    UNKNOWN = "Unknown"



from dataclasses import dataclass
from datetime import datetime

from dataclasses_json import dataclass_json

from .enum import Characteristic, Difficulty
from .fields import datetime_field, default, characteristic_field, difficulty_field


@dataclass_json
@dataclass
class Score:
    """Score data from ScoreSaber"""

    rank: int = default()
    score_id: int = default("scoreId")
    score: int = default()
    unmodified_score: int = default("unmodififiedScore")
    mods: str = default()
    pp: float = default()
    weight: float = default()
    time_set: datetime = datetime_field("timeSet")
    leaderboard_id: int = default("leaderboardId")
    song_hash: str = default("songHash")
    song_name: str = default("songName")
    song_sub_name: str = default("songSubName")
    song_author_name: str = default("songAuthorName")
    level_author_name: str = default("levelAuthorName")
    characteristic: Characteristic = characteristic_field("difficultyRaw")
    difficulty: Difficulty = difficulty_field("difficulty")
    max_score: int = default("maxScore")



import logging

from .scoresaber import ScoreSaber
from .scoresaber_provider import ScoreSaberProvider
from .errors import *
from .models import *

handler = logging.StreamHandler()
logging.getLogger("PyScoreSaber").addHandler(handler)



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="PyScoreSaber",
    version="0.1.1",
    author="LuCkEr-",
    author_email="lucker@lucker.xyz",
    description="Score Saber API client",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/Kiyomi-Parents/PyScoreSaber",
    project_urls={
        "Bug Tracker": "https://github.com/Kiyomi-Parents/PyScoreSaber/issues",
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    package_dir={"": "src"},
    packages=setuptools.find_packages(where="src"),
    python_requires=">=3.8",
    install_requires=[
        "aiohttp",
        "python-dateutil",
        "OutCache",
        "Faker",
        "dataclasses-json",
        "marshmallow"
    ]
)



from dataclasses import dataclass

from dataclasses_json import dataclass_json

from .fields import default


@dataclass_json
@dataclass
class Badge:
    image: str = default()
    description: str = default()
